## 3D Rendering

SpriteJS<super>Next</super> provides [3D extension](https://github.com/spritejs/sprite-extend-3d), which 3D models can be loaded and rendered.

### Rendering 3D models

SpriteJS<super>Next</super> 3D extension based on [OGL](https://github.com/oframe/ogl)。

Thanks to OGL, we can easily render 3D models and interact with them.

```js
const {Scene} = spritejs;
const {Mesh3d, shaders} = spritejs.ext3d;
const container = document.getElementById('container');
const scene = new Scene({
  container,
  displayRatio: 2,
});
const layer = scene.layer3d('fglayer', {
  camera: {
    fov: 35,
  },
  directionalLight: [0.5, 1.0, -0.3],
  directionalLightColor: [1, 1, 1, 0.15],
});

layer.camera.attributes.pos = [8, 5, 15];
layer.camera.lookAt([0, 1.5, 0]);

const texture = layer.createTexture('https://p3.ssl.qhimg.com/t01d6c6c93fdddf1e42.jpg');
const program = layer.createProgram({
  ...shaders.BASE_TEXTURE,
  texture,
});

const model = layer.loadModel('https://s5.ssl.qhres2.com/static/1eb3e9b91a296abd.json');
const fox = new Mesh3d(program, {model});
layer.append(fox);
layer.setOrbit();
```

<iframe src="/demo/#/3d/fox" height="450"></iframe>

### Using geometry elements

We can also easily use Geometry classes to create geometric elements.

Currently supported elements as follows:

- Cube
- Cylinder
- Plane
- Sphere

```js
const {Scene} = spritejs;
const {Cylinder, Sphere, Cube, shaders} = spritejs.ext3d;
const container = document.getElementById('container');
const scene = new Scene({
  container,
  displayRatio: 2,
});
const layer = scene.layer3d('fglayer', {
  directionalLight: [1, 0, 0],
  directionalLightColor: [1, 1, 1, 0.5],
  pointLightColor: `hsl(${Math.floor(360 * Math.random())}, 50%, 50%)`,
  pointLightPosition: [5, 3, 6],
  camera: {
    fov: 35,
  },
});

layer.camera.attributes.pos = [5, 3, 6];
layer.camera.lookAt([0, 0, 0]);

const program = layer.createProgram({
  ...shaders.BASE_GEOMETRY,
  cullFace: null,
  uniforms: {
    lighting: {value: [0.3, 0.8, 0.6, 0.1]},
  },
});

const cylinder = new Cylinder(program);
cylinder.attributes.pos = [0, 1.3, 0];
layer.append(cylinder);
cylinder.animate([
  {rotateY: 0},
  {rotateY: -360},
], {
  duration: 10000,
  iterations: Infinity,
});

const sphere = new Sphere(program);
sphere.attr({
  phiLength: Math.PI,
});
layer.append(sphere);
sphere.animate([
  {rotateY: 0},
  {rotateY: -360},
], {
  duration: 7500,
  iterations: Infinity,
});

const cube = new Cube(program);
cube.attributes.pos = [0, -1.3, 0];
layer.append(cube);
cube.animate([
  {rotateY: 0},
  {rotateY: -360},
], {
  duration: 5000,
  iterations: Infinity,
});

layer.setRaycast();

layer.addEventListener('click', (evt) => {
  if(evt.target === cube) {
    const colors = [];
    for(let i = 0; i < 3; i++) {
      const randomColor = `hsl(${Math.floor(360 * Math.random())}, 50%, 50%)`;
      colors.push(randomColor, randomColor);
    }
    evt.target.attributes.colors = colors;
  } else if(evt.target !== layer) {
    evt.target.attributes.colors = `hsl(${Math.floor(360 * Math.random())}, 50%, 50%)`;
  }
});
```

Try clicking on the geometry elements with your mouse:

<iframe src="/demo/#/3d/geometry" height="750"></iframe>

For more details, see [ext3d](/en/ext3d/index).